#
#  FILE NAME:
#      enum.py
#
#  AUTHOR(S):
#      Dariusz Walczak <dariusz.walczak@gmail.com>
#

# Standard library imports:
import logging

# P3 imports:
from p3.base.exception import BaseModuleError
from p3.base.string_helpers import is_identifier_string, is_keyword_string, safe_fmt, uni_str


_LOG = logging.getLogger( "p3.base" )


class Enum(object):
    def __init__( self, *args ):
        super(Enum, self).__setattr__( "items", {} )
        super(Enum, self).__setattr__( "direct_items", {} )
        super(Enum, self).__setattr__( "reference_items", {} )

        next_value = 0

        for arg in args:
            if isinstance(arg, basestring):
                name, value = arg, next_value
            else:
                try:
                    name, value = arg
                except ValueError as e:
                    _LOG.error( "Can't unpack non string element '%s' into name and value"
                                " variables.", arg )
                    raise

            name = uni_str( name )

            if not is_identifier_string( name ):
                raise BaseModuleError(
                    BaseModuleError.codes.enum_definition_error,
                    safe_fmt( "'%s' enum value name is not a valid identifier.", name ) )
            if is_keyword_string( name ):
                raise BaseModuleError(
                    BaseModuleError.codes.enum_definition_error,
                    safe_fmt( "'%s' enum value name is a python keyword.", name ) )
            if name in self.items:
                raise BaseModuleError(
                    BaseModuleError.codes.enum_definition_error,
                    safe_fmt( "'%s' enum value name is used more than once.", name ) )

            if isinstance(value, basestring):
                self.reference_items[name] = value
            else:
                self.direct_items[name] = value
                self.items[name] = value
                self.__dict__[name] = value
                next_value = value + 1

        for referer, referee in self.reference_items.items():
            self.items[referer] = self._resolve_reference( referer )

        super(Enum, self).__setattr__( "values", tuple(set(self.direct_items.values())) )


    # ===[ Utilities ]=========================================================================== #

    def get_names( self, value, include_references=False ):
        return tuple(self._lookup_names( value, include_references ))


    def get_names_as_string( self, value, include_references=False ):
        l = self._lookup_names( value, include_references )
        l.sort()
        return ','.join( l )


    def get_values( self ):
        return self.values


    # ===[ Item access ]========================================================================= #

    def __contains__( self, element ):
        if isinstance( element, basestring ):
            return element in self.items
        else:
            return element in self.values


    def __getattr__( self, name ):
        if name in self.items:
            return self.items[name]
        else:
            raise AttributeError( name )


    def __setattr__( self, name, value ):
        raise BaseModuleError(
            BaseModuleError.codes.enum_const_error,
            "Enum object doesn't support direct attribute assignments." )


    def __delattr__( self, name ):
        raise BaseModuleError(
            BaseModuleError.codes.enum_const_error,
            "Enum object doesn't support attribute deletion." )


    def __getitem__( self, name ):
        if not isinstance( name, basestring ):
            raise TypeError( safe_fmt( "Key '%s' is of inproper type %s", name, type(name).__name__ ) )

        name = uni_str( name )
        

        if name in self.items:
            return self.items[name]
        else:
            raise IndexError( "Item '%s' not found.", name )


    def __setitem__( self, name, value ):
        raise BaseModuleError(
            BaseModuleError.codes.enum_const_error,
            "Enum item value can't be changed." )


    def __delitem__( self, name ):
        raise BaseModuleError(
            BaseModuleError.codes.enum_const_error,
            "Enum item can't be deleted." )


    def __len__( self ):
        return len(self.items)


    # ===[ Internals ]=========================================================================== #

    def _lookup_names( self, value, include_references ):
        items = (self.items.items() if include_references else self.direct_items.items() )

        return [n for n, v in items if v==value]


    def _resolve_reference( self, referer ):
        ref_trace = []
        init_ref = referer

        def resolve( curr_ref ):
            ref_trace.append( curr_ref )

            if curr_ref in self.direct_items:
                return self.direct_items[curr_ref]
            elif curr_ref in self.reference_items:
                next_ref = self.reference_items[curr_ref]

                if next_ref == curr_ref:
                    raise BaseModuleError(
                        BaseModuleError.codes.enum_definition_error,
                        safe_fmt( "Item '%s' refers itself.", next_ref ) )
                elif next_ref in ref_trace:
                    raise BaseModuleError(
                        BaseModuleError.codes.enum_definition_error,
                        safe_fmt( "Cyclic reference detected: '%s' reference resolves to itself"
                                 " through %s.", init_ref, ", ".join( ref_trace + [next_ref] ) ) )
                else:
                    return resolve( next_ref )
            else:
                raise BaseModuleError(
                    BaseModuleError.codes.enum_definition_error,
                    safe_fmt( "Unresolved '%s' reference to '%s'.", ref_trace[-1], curr_ref ) )

        return resolve( referer )
                    
