#
#  FILE NAME:
#      unique.py
#
#  AUTHOR(S):
#      Dariusz Walczak <dariusz.walczak@gmail.com>
#

# Standard library imports:
import logging

# P3 imports:
from p3.data.error import DataModuleError
from p3.data.idx.base import BaseIndex


_LOG = logging.getLogger( "p3.data" )


class UniqueIndex(BaseIndex):
    def __init__( self, table, *index_cols ):
        super( UniqueIndex, self ).__init__( table )
        self._index_cols = table.resolve_columns( *index_cols )
        self._rows = {}

    def make_key( self, row ):
        cells = row.cells # the property getter constructs a new tuple
        return tuple(cells[c.index].value for c in self._index_cols)


    def _format_key_info( self, row ):
        cells = row.cells # the property getter constructs a new tuple
        def vfmt( col ):
            if col.name is None:
                return "'{0:s}'".format( str(cells[col.index].value) )
            else:
                return "'{0:s}'='{1:s}'".format( col.name, str(cells[col.index].value) )

        return ', '.join( vfmt(c) for c in self._index_cols )


    def get( self, *values ):
        assert len(values) == len(self._index_cols)

        try:
            return (self._rows[values],)
        except KeyError as e:
            return ()


    def index_rows( self, *rows ):
        # Generate items for all indexed rows:
        new_rows = {}

        for row in rows:
            assert row.table is self._table

            key = self.make_key( row )
            if (key in self._rows) or (key in new_rows):
                raise DataModuleError(
                    DataModuleError.codes.row_not_unique,
                    "Uniqueness check failed ({0:s}).".format( self._format_key_info(row) ) )
            else:
                new_rows[key] = row

        # Uniqueness is assured now - merge new_rows with _rows:
        self._rows.update( new_rows )


    def forget_all( self ):
        self._rows.clear()


    def forget_rows( self, *rows ):
        keys = []

        for row in rows:
            assert row.table is self._table

            key = self.make_key( row )
            if key in self._rows:
                keys.append( key )
            else:
                raise DataModuleError(
                    DataModuleError.codes.row_not_found,
                    "Row ({0:s}) is not indexed.".format( self._format_key_info(row) ) )

        for key in keys:
            del self._rows[key]
