#
#  FILE NAME:
#      connection.py
#
#  AUTHOR(S):
#      Dariusz Walczak <dariusz.walczak@gmail.com>
#

# Standard library imports:
import logging

# P3 imports:
from p3.base import safe_fmt
from p3.db.dbms import DBMS, get_dbms_id_by_name
from p3.db.error import DbModuleError


_LOG = logging.getLogger( "p3.db" )


class DbConnection( object ):
    @classmethod
    def create_from_setting( cls, settings, setting_name ):
        db_setting = settings.get_setting( setting_name )
        if db_setting is None:
            raise DbModuleError( DbModuleError.codes.invalid_db_setting,
                                 safe_fmt( "Couldn't find following database connection setting:"
                                           " %s", setting_name ) )

        return cls.create_from_dict( db_setting )

    ## Create DbConnection object from standard connection specification dictionary. Following
    #  dictionary items are required: 'dbms' - database management system name (translated by
    #  the get_dbms_id_by_name function to values of the DBMS enum; 'name' - database name
    #  (optional in case of MySQL); 'user' - database user name. Optional parameters include:
    #  'host' - database server host; 'port' - database server port; 'pass' - database password.
    @classmethod
    def create_from_dict( cls, d ):
        dbms_name = d.get( 'dbms' )
        db_name   = d.get( 'name' )
        db_user   = d.get( 'user' )

        if dbms_name is None:
            raise DbModuleError( DbModuleError.codes.invalid_db_setting,
                                 "Invalid connection specification: missing 'dbms' item (database"
                                 " management system name)." )

        dbms = get_dbms_id_by_name(dbms_name)

        if (db_name is None) and (dbms <> DBMS.mysql):
            raise DbModuleError( DbModuleError.codes.invalid_db_setting,
                                 "Invalid connection specification: missing 'name' item (database"
                                 " name)." )
        if db_user is None:
            raise DbModuleError( DbModuleError.codes.invalid_db_setting,
                                 "Invalid connection specification: missing 'user' item (database"
                                 " user name)." )

        spec = {
            'db_host': d.get( 'host' ),
            'db_port': d.get( 'port' ),
            'db_pass': d.get( 'pass' ) }

        return cls( dbms, db_name, db_user, **spec )


    def __init__( self, dbms, db_name, db_user, db_host=None, db_port=None, db_pass=None ):
        assert dbms in DBMS
        self.dbms    = dbms
        self.db_name = db_name
        self.db_user = db_user
        self.db_host = db_host
        self.db_port = db_port
        self.db_pass = db_pass
        self.db_conn = None
        self.db_cursor = None


    def connect( self ):
        if self.dbms == DBMS.mysql:
            self.db_conn = self._my_sql_connect()
        elif self.dbms == DBMS.pgsql:
            self.db_conn = self._pg_sql_connect()
        else:
            raise DbModuleError( DbModuleError.codes.dbms_not_supported,
                                 safe_fmt( "Unsupported database management system: '%s'",
                                           '/'.join( DBMS.get_names( self.dbms ) ) ) )

    def cursor( self, recreate=False ):
        if self.db_conn is None:
            self.connect()

        if self.db_cursor is not None and not recreate:
            return self.db_cursor
        else:
            _LOG.debug( "Creating new database cursor." )

        if self.dbms == DBMS.mysql:
            import MySQLdb
            self.db_cursor = self.db_conn.cursor( MySQLdb.cursors.DictCursor )
        elif self.dbms == DBMS.pgsql:
            import psycopg2
            import psycopg2.extras
            self.db_cursor = self.db_conn.cursor( cursor_factory=psycopg2.extras.DictCursor )
        else:
            raise DbModuleError( DbModuleError.codes.dbms_not_supported,
                                 safe_fmt( "Unsupported database management system: '%s'",
                                           '/'.join( DBMS.get_names( self.dbms ) ) ) )
        return self.db_cursor

    def release_cursor( self ):
        self.db_cursor = None


    def get_error_class( self ):
        if self.dbms == DBMS.mysql:
            import MySQLdb
            return MySQLdb.Error
        if self.dbms == DBMS.pgsql:
            import psycopg2
            return psycopg2.Error
        else:
            msg = safe_fmt( "Unsupported database management system: '%s'",
                            '/'.join( DBMS.get_names( self.dbms ) ) )
            raise DbModuleError( DbModuleError.codes.dbms_not_supported, msg )


    def get_warning_class( self ):
        if self.dbms == DBMS.mysql:
            import MySQLdb
            return MySQLdb.Warning
        if self.dbms == DBMS.pgsql:
            import psycopg2
            return psycopg2.Warning
        else:
            msg = safe_fmt( "Unsupported database management system: '%s'",
                            '/'.join( DBMS.get_names( self.dbms ) ) )
            raise DbModuleError( DbModuleError.codes.dbms_not_supported, msg )


    def get_executed_query( self ):
        if self.db_cursor is None:
            raise DbModuleError( DbModuleError.codes.cursor_not_used, "Cursor wasn't created." )

        if self.dbms == DBMS.mysql:
            if not hasattr( self.db_cursor, "_last_executed" ):
                raise DbModuleError( DbModuleError.codes.cursor_not_used,
                                     "No query was executed with the current cursor." )
            return self.db_cursor._last_executed
        elif self.dbms == DBMS.pgsql:
            if self.db_cursor.query is None:
                raise DbModuleError( DbModuleError.codes.cursor_not_used,
                                     "No query was executed with the current cursor." )
            return self.db_cursor.query
        else:
            msg = safe_fmt( "Unsupported database management system: '%s'",
                            '/'.join( DBMS.get_names( self.dbms ) ) )
            raise DbModuleError( DbModuleError.codes.dbms_not_supported, msg )


    def build_args( self, mandatory, optional ):
        for n, v in mandatory:
            if v is None:
                raise DbModuleError( DbModuleError.codes.mandatory_arg_not_specified, n )

        return dict(mandatory + tuple((n,v) for n,v in optional if v is not None))

    def _my_sql_connect( self ):
        import MySQLdb

        mandatory = (
            ('user', self.db_user),)

        optional = (
            ('db', self.db_name),
            ('host', self.db_host),
            ('port', self.db_port),
            ('passwd', self.db_pass))
        return MySQLdb.connect( **self.build_args( mandatory, optional ) )


    def _pg_sql_connect( self ):
        import psycopg2

        mandatory = (
            ('database', self.db_name),
            ('user',     self.db_user))

        optional = (
            ('password', self.db_pass),
            ('host',     self.db_host),
            ('port',     self.db_port))

        return psycopg2.connect( **self.build_args( mandatory, optional ) )
