#
#  FILE NAME:
#      format.py
#
#  AUTHOR(S):
#      Dariusz Walczak <dariusz.walczak@gmail.com>
#

# Standard library imports:
import inspect
import logging

# Third party imports:
import django.template

# P3 libraries:
from p3.base.import_helpers import import_object

_LOG = logging.getLogger( "p3.web" )
register = django.template.Library()


class FormatNode( django.template.Node ):
    def __init__( self, variable, cb, *cb_args ):
        assert callable(cb)

        self.variable = django.template.Variable( variable )
        self.cb       = cb
        self.cb_args   = cb_args

    def render( self, context ):
        try:
            value = self.variable.resolve( context )
        except django.template.VariableDoesNotExist as e:
            _LOG.error( "Exception catched: %s", str(e) )
            raise
        return self.cb( value, *self.cb_args )


# format tag syntax:
# {% format <variable> <callback-path> [<callback-args> [, ...]] %}
#
# TestCase1:  {% format %} ERROR
# TestCase2:  {% format var %} ERROR
# TestCase3:  {% format "value" %} ERROR
# TestCase4:  {% format p3.some_path.SomeFun %} ERROR
# TestCase5:  {% format var p3.some_path.SomeFun %} OK
# TestCase6:  {% format "value" p3.some_path.SomeFun %} OK?
# TestCase5:  {% format var p3.some_path.NOT_CALLABLE %} ERROR
# TestCase7:  {% format var p3.invalid_path.SomeFun %} ERROR
# TestCase8:  {% format var p3.some_path.SomeFun "some_expected_arg" %} OK
# TestCase9:  {% format var p3.some_path.SomeFun "some_expected_arg" "some_optional_arg" %} OK
# TestCase10: {% format var p3.some_path.SomeFun "some_unexpected_arg" %} ERROR
@register.tag( name="format" )
def do_format( parser, token ):
    _LOG.debug( "Parsing template tag: '%s'", token )

    # Initially parse and verify tag arguments:
    signature = token.split_contents()

    if len(signature) < 3:
        _LOG.debug( "Inproper argument count" )
        raise django.template.TemplateSyntaxError(
            "'{0:s}' tag requires at least two arguments (variable and format callback path)."
            " Following arguments were provided: {1:s}.".format(
                signature[0], ', '.join(signature[1:]) ) )

    variable = signature[1]
    cb_path   = signature[2]
    cb_args  = signature[3:]

    # Attempt to import any of callback candidates:
    CB_CANDIDATES = (
        cb_path,
        "{0:s}.{1:s}".format( "p3.web.django_apps.template_ext", cb_path ))

    cb = None

    for path in CB_CANDIDATES:
        try:
            cb = import_object( path )
            break;
        except ImportError as e:
            _LOG.debug( "Couldn't import '%s' callback. Exception catched: %s",
                        path, str(e) )

    if cb is None:
        msg = "Couldn't import callback from any of following candidates: {0:s}".format(
            ', '.join( "'{0:s}'".format(p) for p in CB_CANDIDATES) )
        _LOG.error( msg )
        raise django.template.TemplateSyntaxError( msg )

    # Verify imported callback:
    if not callable(cb):
        msg = "Object '{0:s}' is not callable.".format( cb_path )
        _LOG.error( msg )
        raise django.template.TemplateSyntaxError( msg )

    # Check if provided arguments match callback's signature:
    arg_spec = inspect.getargspec(cb)
    opt_arg_cnt = (0 if arg_spec.defaults is None else len(arg_spec.defaults))
    req_arg_cnt = len(arg_spec.args)-1-opt_arg_cnt
    has_var_arg = arg_spec.varargs is not None

    if len(cb_args) < req_arg_cnt:
        msg = "Format callback '{0:s}' requires at least {1:d} additional arguments ({2:d} were"\
            " provided).".format( cb_path, req_arg_cnt, len(cb_args) )
        raise django.template.TemplateSyntaxError( msg )

    if not has_var_arg and (len(cb_args) > req_arg_cnt+opt_arg_cnt):
        msg = "Format callback '{0:s}' expects at most {1:d} additional arguments ({2:d} were"\
            " provided).".format( cb_path, req_arg_cnt, len(cb_args) )
        raise django.template.TemplateSyntaxError( msg )

    # Finally, construct the node:
    return FormatNode( variable, cb, *cb_args )
